from __future__ import absolute_import
from builtins import str
from builtins import object
import numpy as np

from .iaugmentor import IAugmentor
from .utils.readCOCOJSON import readCOCOJSON
import json
import cv2
from sklearn.externals.joblib import Parallel, delayed
import imutils


def readAndGenerateInstanceSegmentation(outputPath, transformers, inputPath, imageInfo, annotationsInfo):
    name = imageInfo[0]
    imagePath = inputPath + "/" + name
    (w, h) = imageInfo[1]
    image = cv2.imread(imagePath)
    maskLabels = []
    for (c, annotation) in annotationsInfo:
        mask = np.zeros((w, h), dtype="uint8")
        annotation = [[annotation[2 * i], annotation[2 * i + 1]] for i in range(0, int(len(annotation) / 2))]
        pts = np.array([[int(x[0]), int(x[1])] for x in annotation], np.int32)
        pts = pts.reshape((-1, 1, 2))
        cv2.fillPoly(mask, [pts], True, 255)
        maskLabels.append((mask, c))

    allNewImagesResult = []
    for (j, transformer) in enumerate(transformers):
        (newimage, newmasklabels) = transformer.transform(image, maskLabels)

        cv2.imwrite(outputPath + str(j) + "_" + name, newimage)
        newSegmentations = []
        for (mask, label) in newmasklabels:
            cnts = cv2.findContours(mask.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
            cnts = cnts[0] if imutils.is_cv2() else cnts[1]
            segmentation = [[x[0][0], x[0][1]] for x in cnts[0]]
            # Closing the polygon
            segmentation.append(segmentation[0])

            newSegmentations.append((label, cv2.boundingRect(cnts[0]), segmentation, cv2.contourArea(cnts[0])))

        allNewImagesResult.append((str(j) + "_" + name, (w, h), newSegmentations))

    return allNewImagesResult


# This class serves to generate images for an instance segmentation
# problem where all the images are organized in a folder called
# images and there is a json file with the annotations of the images
# using the COCO format called annotation.json

class COCOLinearInstanceSegmentationAugmentor(IAugmentor):

    def __init__(self, inputPath, parameters):
        IAugmentor.__init__(self)
        self.imagesPath = inputPath
        self.annotationFile = inputPath + "/annotations.json"
        # output path represents the folder where the images will be stored
        if parameters["outputPath"]:
            self.outputPath = parameters["outputPath"]
        else:
            raise ValueError("You should provide an output path in the parameters")


    def readImagesAndAnnotations(self):
        (self.info, self.licenses, self.categories, self.dictImages, self.dictAnnotations) \
            = readCOCOJSON(self.annotationFile)

    def applyAugmentation(self):
        self.readImagesAndAnnotations()

        newannotations = Parallel(n_jobs=-1)(delayed(readAndGenerateInstanceSegmentation)
                                             (self.outputPath, self.transformers, self.imagesPath, self.dictImages[x],
                                              self.dictAnnotations[x])
                                             for x in self.dictImages.keys())

        data = {}
        data['info'] = self.info
        data['licenses'] = self.licenses
        data['categories'] = self.categories
        data['images'] = []
        data['annotations'] = []
        imageId = 1
        annotationId = 1
        newannotations = [item for sublist in newannotations for item in sublist]
        for (fil, (w, h), annotations) in newannotations:
            data['images'].append({'id': imageId,
                                   'file_name': fil,
                                   'width': w,
                                   'height': h,
                                   'date_captured': '',
                                   'license': 1,
                                   'coco_url': '',
                                   'flickr_url': ''})

            for annotation in annotations:

                label = annotation[0]
                rect = annotation[1]
                segmentation = annotation[2]
                area = annotation[3]
                segmentationCOCO = []
                for x in segmentation:
                    segmentationCOCO.append(int(x[0]))
                    segmentationCOCO.append(int(x[1]))
                data['annotations'].append({'id': annotationId,
                                            'image_id': imageId,
                                            'category_id': label,
                                            'iscrowd': 0,
                                            'area': area,
                                            'bbox': [rect[0], rect[1], rect[2], rect[3]],
                                            'segmentation': [segmentationCOCO],
                                            'width': w,
                                            'height': h})
                annotationId += 1
            imageId += 1

        with open(self.outputPath + "annotation.json", 'w') as outfile:
            json.dump(data, outfile)




from abc import ABCMeta, abstractmethod
from future.utils import with_metaclass


class Transformer(with_metaclass(ABCMeta, object)):
    def __init__(self, technique,dictLabels=None):
        self.technique = technique
        self.dictLabels=dictLabels

    @abstractmethod
    def transform(self, T1,T2):
        raise NotImplementedError

    def transformLabel(self,label):
        if (self.dictLabels is not None) and (label in self.dictLabels.keys()):
            return self.dictLabels[label]
        else:
            return label




from __future__ import absolute_import
from .transformer import Transformer
from .detection import detectBoxes
from ..techniques.technique import PositionVariantTechnique



class TransformerForImageDetection(Transformer):

    def __init__(self,technique,dictLabels=None):
        Transformer.__init__(self,technique,dictLabels)

    def transform(self, image, boxes):
        newImage = self.technique.apply(image)

        if (isinstance(self.technique, PositionVariantTechnique)):
            newBoxes = detectBoxes(image.shape[:2], boxes, self.technique)
            newBoxes = [(self.transformLabel(box[0]),box[1]) for box in newBoxes]
        else:
            newBoxes= [(self.transformLabel(box[0]),box[1]) for box in boxes]

        return [newImage,newBoxes]


from __future__ import absolute_import
from builtins import str
from builtins import object
from .iaugmentor import IAugmentor
from imutils import paths
import os
import cv2
from sklearn.externals.joblib import Parallel, delayed

# We need to define this function outside to work in parallel.
def readAndGenerateVideo(outputPath, transformers, i_and_videoPath):
    (i, videoPath) = i_and_videoPath

    vidcap = cv2.VideoCapture(videoPath)
    fps = vidcap.get(cv2.CAP_PROP_FPS)
    listImages = []
    success, image = vidcap.read()
    (h,w) = image.shape[:2]
    listImages.append(image)
    while success:
        success, image = vidcap.read()
        listImages.append(image)
    vidcap.release()

    label = videoPath.split(os.path.sep)[-2]
    name = videoPath.split(os.path.sep)[-1]
    for (j, transformer) in enumerate(transformers):
        newListImages,newlabel = transformer.transform(listImages,label)

        out = cv2.VideoWriter(outputPath + newlabel + "/" + str(i) + "_" + str(j) + "_" + name,
                          cv2.VideoWriter_fourcc('M', 'J', 'P', 'G'), fps, (w, h))

        for image in newListImages:
            out.write(image)

        out.release()

# This class serves to generate videos for a classification
# problem where all the videos are organized by folders
# distributed by labels. Example:
# - Folder
# |- cats
#    |- video1.mp4
#    |- video2.mp4
#    |- ...
# |- dogs
#    |- video1.mp4
#    |- video2.mp4
#    |- ...
class FolderVideoLinearClassificationAugmentor(IAugmentor):

    def __init__(self,inputPath,parameters):
        IAugmentor.__init__(self)
        self.inputPath = inputPath
        # output path represents the folder where the images will be stored
        if parameters["outputPath"]:
            self.outputPath = parameters["outputPath"]
        else:
            raise ValueError("You should provide an output path in the parameters")




    def readImagesAndAnnotations(self):
        self.imagePaths = list(paths.list_files(self.inputPath,validExts=(".avi", ".mp4")))

    def createOutputDirs(self):
        dirs = [d for d in os.listdir(self.inputPath) if os.path.isdir(os.path.join(self.inputPath, d))]
        for dir in dirs:
            os.makedirs(self.outputPath + dir + "/")

    def applyAugmentation(self):
        self.readImagesAndAnnotations()
        self.createOutputDirs()

        #[readAndGenerateVideo(self.outputPath,self.generators,x) for x in enumerate(self.imagePaths)]
        Parallel(n_jobs=-1)(delayed(readAndGenerateVideo)(self.outputPath,self.transformers,x) for x in enumerate(self.imagePaths))




from __future__ import absolute_import
from builtins import str
from builtins import object
from .iaugmentor import IAugmentor
from imutils import paths
import os
import cv2
from sklearn.externals.joblib import Parallel, delayed

# We need to define this function outside to work in parallel.
def readAndGenerateImage(outputPath, transformers, i_and_imagePath):
    (i, imagePath) = i_and_imagePath
    image = cv2.imread(imagePath)
    label = imagePath.split(os.path.sep)[-2]
    name = imagePath.split(os.path.sep)[-1]
    for (j, transformer) in enumerate(transformers):
        newimage,newlabel = transformer.transform(image,label)
        cv2.imwrite(outputPath + newlabel + "/" + str(i) + "_" + str(j) + "_" + name,
                    newimage)

# This class serves to generate images for a classification
# problem where all the images are organized by folders
# distributed by labels. Example:
# - Folder
# |- cats
#    |- image1.jpg
#    |- image2.jpg
#    |- ...
# |- dogs
#    |- image1.jpg
#    |- image2.jpg
#    |- ...
class FolderLinearClassificationAugmentor(IAugmentor):

    def __init__(self,inputPath,parameters):
        IAugmentor.__init__(self)
        self.inputPath = inputPath
        # output path represents the folder where the images will be stored
        if parameters["outputPath"]:
            self.outputPath = parameters["outputPath"]
        else:
            raise ValueError("You should provide an output path in the parameters")

    def readImagesAndAnnotations(self):
        self.imagePaths = list(paths.list_files(self.inputPath,validExts=(".jpg", ".jpeg", ".png", ".bmp",".tiff",".tif")))

    def createOutputDirs(self):
        dirs = [d for d in os.listdir(self.inputPath) if os.path.isdir(os.path.join(self.inputPath, d))]
        for dir in dirs:
            os.makedirs(self.outputPath + dir + "/")

    def applyAugmentation(self):
        self.readImagesAndAnnotations()
        self.createOutputDirs()
        len1 = len(self.imagePaths)
        Parallel(n_jobs=-1)(delayed(readAndGenerateImage)(self.outputPath,self.transformers,x) for x in enumerate(self.imagePaths))




import os
import cv2
import xml.etree.ElementTree as ET
import argparse

ap = argparse.ArgumentParser()
ap.add_argument("-i", "--image", required=True, help="path to the image")
ap.add_argument("-l", "--label", required=True, help="path to the label")
args = vars(ap.parse_args())

image = cv2.imread(args["image"])
label = cv2.imread(args["label"])

blending = cv2.addWeighted(image,1,label,0.2,0.0)
# cv2.imshow("Blending",blending)
# cv2.waitKey(0)
cv2.imwrite("blending3.jpg",blending)


from __future__ import absolute_import
from .transformer import Transformer
from ..techniques.technique import PositionVariantTechnique



class TransformerForImageSegmentation(Transformer):

    def __init__(self,technique,dictLabels=None):
        Transformer.__init__(self,technique,dictLabels)

    def transform(self, image, imageLabel):
        if (isinstance(self.technique, PositionVariantTechnique)):
            return [self.technique.apply(image),
                    self.technique.apply(imageLabel)]
        else:
            return [self.technique.apply(image), imageLabel]


from __future__ import absolute_import
from .technique import PositionVariantTechnique
import cv2
import numpy as np

class translationAugmentationTechnique(PositionVariantTechnique):

    def __init__(self,parameters):
        PositionVariantTechnique.__init__(self, parameters)
        if 'x' in list(parameters.keys()):
            self.x = int(parameters["x"])
        else:
            self.x = 10
        if 'y' in list(parameters.keys()):
            self.y = int(parameters["y"])
        else:
            self.y = 10

    def __translate(self,image, x, y):
        # define the translation matrix and perform the translation
        M = np.float32([[1, 0, x], [0, 1, y]])
        shifted = cv2.warpAffine(image, M, (image.shape[1], image.shape[0]))

        # return the translated image
        return shifted

    def apply(self, image):
        translation = self.__translate(image, self.x,self.y)
        return translation

# technique = translationAugmentationTechnique(5,5)
# image = cv2.imread("LPR1.jpg")
# cv2.imshow("t",technique.applyForClassification(image))
# cv2.waitKey(0)



import os
import cv2
import xml.etree.ElementTree as ET
import argparse

ap = argparse.ArgumentParser()
ap.add_argument("-i", "--image", required=True, help="path to the image, there must be also an xml file with the same name (but with xml extension)")
args = vars(ap.parse_args())



def readAndShowAnnotatedImage(imagePath):

    image = cv2.imread(imagePath)
    name = imagePath.split(os.path.sep)[-1]
    labelPath = '/'.join(imagePath.split(os.path.sep)[:-1]) + "/"+name[0:name.rfind(".")] + ".xml"
    tree = ET.parse(labelPath)
    root = tree.getroot()
    objects = root.findall('object')
    boxes = []
    for object in objects:
        category = object.find('name').text
        bndbox = object.find('bndbox')
        x  = int(bndbox.find('xmin').text)
        y = int(bndbox.find('ymin').text)
        w = int(bndbox.find('ymax').text)-y
        h = int(bndbox.find('xmax').text) - x
        boxes.append((category, (x, y, w, h)))

    for box in enumerate(boxes):
        box = box[1][1]
        cv2.rectangle(image,(box[0],box[1]),(box[0]+box[3],box[1]+box[2]),(0,0,255),2)

    cv2.imshow("Annotated image",image)
    cv2.waitKey(0)
    # cv2.imwrite("im.jpg",image)

readAndShowAnnotatedImage(args["image"])


from __future__ import absolute_import
from .transformer import Transformer



class TransformerForImageClassification(Transformer):

    def __init__(self,technique,dictLabels=None):
        Transformer.__init__(self,technique,dictLabels)


    def transform(self, image,label):
        newImage = self.technique.apply(image)
        newLabel = self.transformLabel(label)
        return (newImage,newLabel)





from builtins import object
from keras.models import  Sequential
from keras.layers.normalization import BatchNormalization
from keras.layers.convolutional import Conv2D
from keras.layers.convolutional import MaxPooling2D
from keras.layers.core import Activation
from keras.layers.core import Flatten
from keras.layers.core import Dropout
from keras.layers.core import Dense
from keras import backend as K

class MiniVGGNet(object):
    @staticmethod
    def build(width,height,depth,classes=2):
        model = Sequential()
        inputShape = (height,width,depth)
        chanDim=-1
        if K.image_data_format() == "channels_first":
            inputShape = (depth, height, width)
            chanDim=1

        model.add(Conv2D(32,(3,3),padding="same",
                        input_shape=inputShape))
        model.add(Activation("relu"))
        model.add(BatchNormalization(axis=chanDim))
        model.add(Conv2D(32, (3, 3), padding="same"))
        model.add(Activation("relu"))
        model.add(BatchNormalization(axis=chanDim))
        model.add(MaxPooling2D(pool_size=(2,2)))
        model.add(Dropout(0.25))
        model.add(Conv2D(64, (3, 3), padding="same"))
        model.add(Activation("relu"))
        model.add(BatchNormalization(axis=chanDim))
        model.add(Conv2D(64, (3, 3), padding="same"))
        model.add(Activation("relu"))
        model.add(BatchNormalization(axis=chanDim))
        model.add(MaxPooling2D(pool_size=(2, 2)))
        model.add(Dropout(0.25))
        model.add(Flatten())
        model.add(Dense(512))
        model.add(Activation("relu"))
        model.add(BatchNormalization())
        model.add(Dropout(0.5))
        model.add(Dense(classes))
        model.add(Activation("softmax"))

        return model


from __future__ import absolute_import
from .transformerList import Transformers



def transformerGenerator(problem):
    transformer = Transformers[problem]
    if transformer is None:
        raise ValueError("That type of problem is not available")

    return lambda technique,dictLabels=None : transformer(technique,dictLabels)


from __future__ import print_function
from __future__ import absolute_import
from .augmentors.augmentorFactory import createAugmentor
from .transformers.transformerFactory import transformerGenerator
from .techniques.techniqueFactory import createTechnique
import argparse
from .utils.conf import Conf
import cv2
from .mosaic import generateMosaic


def main(conffile,imfile):
	#ap = argparse.ArgumentParser()
	#ap.add_argument("-c", "--conf", required=True, help="path to configuration file")
	#ap.add_argument("-i", "--image", required=True, help="path to the image to generate the sample")
	#args = vars(ap.parse_args())


	conf = Conf(conffile)
	image = cv2.imread(imfile)

	# First, we read the parameters
	problem = conf["problem"]
	annotationMode = conf["annotation_mode"]
	outputMode = conf["output_mode"]
	generationMode = conf["generation_mode"]
	inputPath = conf["input_path"]
	parameters = conf["parameters"]
	augmentationTechniques = conf["augmentation_techniques"]
	print(augmentationTechniques)
	# Second, we create the augmentor
	augmentor = createAugmentor(problem,annotationMode,outputMode,generationMode,inputPath,
		                    parameters)


	# We apply the augmentation
	images = []
	transformer = transformerGenerator(problem)
	for (technique,parameters) in augmentationTechniques:
	    tech = createTechnique(technique,parameters)
	    im = transformer(tech).transform(image)
	    cv2.putText(im,technique,(10,10),cv2.FONT_HERSHEY_SIMPLEX,0.5,(0,0,255),1)
	    images.append(im)

	cv2.imshow("Mosaic",generateMosaic(images))
	cv2.waitKey(0)




from __future__ import absolute_import
from __future__ import division
from past.utils import old_div
from .technique import PositionInvariantTechnique
import cv2
import numpy as np

class raiseHueAugmentationTechnique(PositionInvariantTechnique):

    # Valid values for pover are in the range (0.25,4]
    def __init__(self,parameters):
        PositionInvariantTechnique.__init__(self, parameters)
        if 'power' in list(parameters.keys()):
            self.power = float(parameters["power"])
        else:
            self.power = 1.5

        if (self.power<=0.25 or self.power >4):
            raise NameError("Invalid value for power")

    def apply(self, image):
        if(len(image.shape)!=3):
            raise NameError("Not applicable technique")
        imageHSV = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        identityV = np.arange(256, dtype=np.dtype('uint8'))
        identityH = np.array([((old_div(i, 255.0)) ** self.power) * 255
                              for i in np.arange(0, 256)]).astype("uint8")
        identityS = np.arange(256, dtype=np.dtype('uint8'))
        lut = np.dstack((identityH, identityS, identityV))

        # apply gamma correction using the lookup table
        imageHSV = cv2.LUT(imageHSV, lut)
        imageRGB = cv2.cvtColor(imageHSV, cv2.COLOR_HSV2BGR)
        return imageRGB


import json




def readCOCOJSON(jsonPath):
    with open(jsonPath) as f:
        data = json.load(f)

    info = data['info']
    licenses = data['licenses']
    categories = data['categories']
    images = data['images']
    annotations = data['annotations']

    dictImages = {}
    dictAnnotations = {}
    for image in images:
        dictImages[image['id']] = (image['file_name'],(image['width'],image['height']))
        dictAnnotations[image['id']] = []

    for annotation in annotations:
        dictAnnotations[annotation['image_id']].append((annotation['category_id'],
                                                        annotation['segmentation'][0]))

    return (info,licenses,categories,dictImages,dictAnnotations)






from __future__ import absolute_import
from .transformer import Transformer
from ..techniques.technique import PositionVariantTechnique



class TransformerForImageInstanceSegmentation(Transformer):

    def __init__(self,technique,dictLabels=None):
        Transformer.__init__(self,technique,dictLabels)

    def transform(self, image,maskLabels):
        if (isinstance(self.technique, PositionVariantTechnique)):
            return [self.technique.apply(image),
                    [(self.technique.apply(mask), self.transformLabel(label))
                     for mask,label in maskLabels ]]
        else:
            return [self.technique.apply(image),
                    [(mask, self.transformLabel(label))
                     for mask, label in maskLabels]]


from builtins import object
from abc import ABCMeta,abstractmethod
from future.utils import with_metaclass

class IAugmentor(with_metaclass(ABCMeta, object)):
    def __init__(self):
        self.transformers = []

    def addTransformer(self, transformer):
        self.transformers.append(transformer)

    @abstractmethod
    def readImagesAndAnnotations(self):
        raise NotImplementedError

    @abstractmethod
    def applyAugmentation(self):
        raise NotImplementedError



from __future__ import absolute_import
from .technique import PositionVariantTechnique
import cv2
import numpy as np

class shearingAugmentationTechnique(PositionVariantTechnique):

    def __init__(self,parameters):
        PositionVariantTechnique.__init__(self, parameters)
        if 'a' in list(parameters.keys()):
            self.a = float(parameters["a"])
        else:
            self.a = 0.5

    def __shear(self,image, a):
        # define the translation matrix and perform the translation
        M = np.float32([[1, a, 0], [0, 1, 0]])
        sheared = cv2.warpAffine(image, M, (image.shape[1], image.shape[0]))

        # return the translated image
        return sheared

    def apply(self, image):
        sheared = self.__shear(image, self.a)
        return sheared

# technique = shearingAugmentationTechnique(0.5)
# image = cv2.imread("LPR1.jpg")
# cv2.imshow("t",technique.applyForClassification(image))
# cv2.waitKey(0)




from __future__ import absolute_import
from .transformerForImageClassification import TransformerForImageClassification
from .transformerForImageDetection import TransformerForImageDetection
from .transformerForImageInstanceSegmentation import TransformerForImageInstanceSegmentation
from .transformerForImageLocalization import TransformerForImageLocalization
from .transformerForImageSegmentation import TransformerForImageSegmentation
from .transformerForImageStackClassification import TransformerForImageStackClassification
from .transformerForImageStackDetection import TransformerForImageStackDetection
from .transformerForImageStackSegmentation import TransformerForImageStackSegmentation


Transformers = {
    "classification" : TransformerForImageClassification,
    "localization" : TransformerForImageLocalization,
    "detection" : TransformerForImageDetection,
    "semantic_segmentation" : TransformerForImageSegmentation,
    "instance_segmentation" : TransformerForImageInstanceSegmentation,
    "stackclassification" : TransformerForImageStackClassification,
    "stackdetection": TransformerForImageStackDetection,
    "stacksegmentation" : TransformerForImageStackSegmentation,
}


from __future__ import absolute_import
from __future__ import division
from past.utils import old_div
from .technique import PositionVariantTechnique
import cv2
import numpy as np
import random

class rotateAugmentationTechnique(PositionVariantTechnique):

    # Valid angle is in the range [0,360), or it can also be a pair
    # indicating the range of angles
    def __init__(self,parameters):
        PositionVariantTechnique.__init__(self, parameters)
        if 'angle' in list(parameters.keys()):
            angle = int(parameters["angle"])
        else:
            angle = random.randint(0,360)

        if (isinstance(angle,list) and len(angle)==2):
            self.angle = random.randint(angle[0],angle[1])
        elif (isinstance(angle,int)):
            self.angle = angle
        else:
            raise ValueError("You must introduce either an integer or a pair")

    def __rotate_bound(self,image):
        # grab the dimensions of the image and then determine the
        # center
        (h, w) = image.shape[:2]
        (cX, cY) = (w // 2, h // 2)

        # grab the rotation matrix (applying the negative of the
        # angle to rotate clockwise), then grab the sine and cosine
        # (i.e., the rotation components of the matrix)
        M = cv2.getRotationMatrix2D((cX, cY), -self.angle, 1.0)
        cos = np.abs(M[0, 0])
        sin = np.abs(M[0, 1])

        # compute the new bounding dimensions of the image
        nW = int((h * sin) + (w * cos))
        nH = int((h * cos) + (w * sin))

        # adjust the rotation matrix to take into account translation
        M[0, 2] += (old_div(nW, 2)) - cX
        M[1, 2] += (old_div(nH, 2)) - cY

        # perform the actual rotation and return the image
        return cv2.warpAffine(image, M, (nW, nH))

    def apply(self, image):
        rotated = self.__rotate_bound(image)
        return rotated


# Example
# technique = rotateAugmentationTechnique(25)
# image = cv2.imread("LPR1.jpg")
# # box = ('dog',(0,0,20,20))
# # imageC=technique.applyForClassification(image)
# # cv2.imshow("classification",technique.applyForClassification(image))
# # (imageL,(_,(x,y,w,h)))=technique.applyForLocalization(image,box)
# # cv2.rectangle(imageL,(x,y),(x+w,y+h),(255,255,255),1)
# cv2.imshow("rotation",technique.apply(image))
# cv2.waitKey(0)